/*
 * $Id: getstream.c,v 1.8 2007/02/10 21:42:48 simimeie Exp $
 *
 */

#include <inttypes.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>

#include <event.h>

#include "getstream.h"
#include "demux.h"
#include "log.h"
#include "dvbrc.h"
#include "fe.h"
#include "crc32.h"
#include "sap.h"
#include "ttyhandlers.h"

struct stream *			streams=NULL;

struct pidentry			pidtable[MAX_PID];

unsigned int			maxuccs = 5; /* (Default) Maximum number of unicast clients */

unsigned int			sndbuf_size = 65536 * 16; /* 1024 KB */

time_t				startupts = 0; /* Startup-Timestamp */

#ifdef COUNTSTATS
uint64_t			totrxbytes = 0; /* Total network counters */
uint64_t			tottxbytes = 0;
unsigned long			totconattempts = 0; /* Connection attempts (only unicast) */
unsigned long			totconsuccesss = 0; /* Successful connections */
uint64_t			totbytesfromadapter = 0; /* Bytes read from adapter */
#endif /* COUNTSTATS */

#define VERSIONSUF "+fox9"


void sigpipehandler(int bla) { /* Dummyhandler for catching the event */
	return; /* It does nothing, but without it our program would die
		 * on receiving a SIGPIPE */
}

void siginthandler(int bla) {
	putlog(0, "Ctrl+C OMG!!!!!!1111");
	jmptocol1();
	cleartoeol();
	exit(0); /* That way our "atexit" function(s) get called */
	return;
}


void send_ts_packet(struct stream * s, uint8_t * tsp) {
	int len;

	/* Copy TS packet to packet buffer */
	memcpy(&s->pbuf[s->pbufvalid], tsp, TS_PACKET_SIZE);
	s->pbufvalid += TS_PACKET_SIZE;

	/* check whether another packet would fit ? */
	if (s->pbufvalid + TS_PACKET_SIZE > MAX_MCAST_PAYLOAD) {
		/* Send packet and reset valid counter */
		if (s->nuccs < 0) {
			len = send(s->sockfd, s->pbuf, s->pbufvalid, MSG_DONTWAIT);
#ifdef COUNTSTATS
			if (len > 0) {
				s->txbytes += len;
				tottxbytes += len;
			}
#endif /* COUNTSTATS */
			if (len != s->pbufvalid) {
				putlog(0, "send didn't send all... %d/%d", len, s->pbufvalid);
			}
		} else {
			/* Send to all registered clients */
			struct unicast_client * uc = s->ucclients;
			while (uc) {
				if (!uc->hadrequest) {
					uc = uc->next;
					continue;
				}
				len = write(uc->sockfd, s->pbuf, s->pbufvalid);
#ifdef COUNTSTATS
				if (len > 0) {
					s->txbytes += len;
					uc->txbytes += len;
					tottxbytes += len;
				}
#endif /* COUNTSTATS */
				if (len != s->pbufvalid) {
					putlog(2, "[fd %05d]: failed send for the %luth time... %d/%d",
						uc->sockfd, uc->sendoverflows, len, s->pbufvalid);
					uc->sendoverflows++;
					if (uc->sendoverflows > MAXSENDOVERFLOWS) {
						struct unicast_client * olduc = uc;
						uc = uc->next;
						putlog(1, "[fd %05d]: Too many overflows, dropping connection.",
							  olduc->sockfd);
						drop_tcp_connection(s, olduc->sockfd);
						continue;
					}
				}
				uc = uc->next;
			}
		}
		s->pbufvalid = 0;
	}
}


int parse_cmdline_program(char *optarg) {
	struct stream		*s;
	char			*pnrend,
				*ipend;
	int			cid;

	s = calloc(1, sizeof(struct stream));

	s->ttl = 1;		/* Set default - only used in SAP currently */

	/*
	 * Parse 0xfd:239.0.1.2:3000
	 * Which is 
	 *
	 * pnr:ipaddr:port
	 * FIXME: This is not a good idea for IPv6...
	 *
	 */

	cid=strtol(optarg, &pnrend, 0);

	if (pnrend == optarg)
		return 0;

	s->cid=cid;

	/* Do we have a pnr ipadr seperator ? */
	if (*pnrend++ != ':')
		return 0;

	ipend=strchr(pnrend, ':');

	if (!ipend) 
		return 0;

	/* Copy ip address, null termination through calloc */
	strncpy(s->ipaddr, pnrend, ipend-pnrend);

	s->port=strtol(ipend+1, NULL, 10);

	s->next=streams;
	streams=s;

	return 1;
}

void setfdnonblocking(int fd) {
	int flags;
	flags = fcntl(fd, F_GETFL, 0);
	flags |= O_NONBLOCK|O_NDELAY;
	if (fcntl(fd, F_SETFL, flags) < 0) {
		perror("nonblocking fcntl");
		exit(1);
	}
}

void drop_tcp_connection(struct stream * s, int fd) {
	struct unicast_client * uc = s->ucclients;
	struct unicast_client * prev = NULL;
	
	while (uc) {
		if (uc->sockfd == fd) { /* That's the one. */
			event_del(&uc->ev);
			close(uc->sockfd);
			if (prev == NULL) {
				s->ucclients = uc->next;
			} else {
				prev->next = uc->next;
			}
			free(uc);
			s->nuccs--;
			return;
		}
		prev = uc;
		uc = uc->next;
	}
}

#define HTTP_OK_REPLY "HTTP/1.0 200 OK\r\n"\
                      "Content-Type: video/mpeg2stream\r\n"\
                      "\r\n"
#define HTTP_OVERLOAD_REPLY "HTTP/1.0 503 Service Unavailable\r\n"\
                            "Content-Type: text/plain\r\n"\
                            "\r\n"\
                            "The server has reached its client limit.\r\n"\
                            "Try again later.\r\n"
void work_tcp_connection(int fd, short event, void * up) {
	struct stream *	s = (struct stream *)up;
	char buf[1024];
	
	if (event & EV_READ) {
		int rdres;
		rdres = read(fd, buf, 1023);
		if (rdres <= 0) {
			putlog(1, "[fd %05d]: Read error, dropping connection", fd);
			drop_tcp_connection(s, fd);
			return;
		} else {
			struct unicast_client * uc = s->ucclients;
			while (uc) {
				if (uc->sockfd == fd) {
#ifdef COUNTSTATS
					totrxbytes += rdres;
					s->rxbytes += rdres;
					uc->rxbytes += rdres;
					tottxbytes += strlen(HTTP_OK_REPLY);
					s->txbytes += strlen(HTTP_OK_REPLY);
					uc->txbytes += strlen(HTTP_OK_REPLY);
#endif /* COUNTSTATS */
					if (!uc->hadrequest) {
						write(uc->sockfd,
						      HTTP_OK_REPLY,
						      strlen(HTTP_OK_REPLY));
						uc->hadrequest = 1;
					}
				}
				uc = uc->next;
			}
#if 0
			buf[rdres]=0;
			printf("Read: %d bytes: %s\n", rdres, buf);
#endif
		}
	}
}

void new_tcp_connection(int fd, short event, void * up) {
	struct stream *	s = (struct stream *)up;
	struct sockaddr_in remaddr;
	unsigned int sl = sizeof(struct sockaddr_in);
	int newsock;
	struct unicast_client *	uc;
	int sopt;
	
	newsock = accept(fd, (struct sockaddr *)&remaddr, &sl);
	if (newsock < 0) {
		perror("(ignoring error) ACCEPT");
		return;
	}
	putlog(1, "[fd %05d]: New connection from %s to stream '%s'",
	          newsock, inet_ntoa(remaddr.sin_addr), s->channel->name);
#ifdef COUNTSTATS
	s->conattempts++;
	totconattempts++;
#endif /* COUNTSTATS */
	setfdnonblocking(newsock);
	if (s->nuccs >= maxuccs) {
		putlog(1, "[fd %05d]: Too many connections already, dropping the new one.",
		          newsock);
		write(newsock, HTTP_OVERLOAD_REPLY, sizeof(HTTP_OVERLOAD_REPLY));
		close(newsock);
		return;
	}
#ifdef COUNTSTATS
	s->consuccesss++;
	totconsuccesss++;
#endif /* COUNTSTATS */
	sopt = sndbuf_size;
	if (setsockopt(newsock, SOL_SOCKET, SO_SNDBUF, (char *)&sopt, sizeof(sopt))) {
		perror("setsockopt SNDBUF");
	}
	
	uc = calloc(1, sizeof(struct unicast_client));
	uc->sockfd = newsock;
	uc->remaddr = remaddr;
	uc->next = s->ucclients;
	s->ucclients = uc;
	s->nuccs++;
	
	event_set(&uc->ev, uc->sockfd, EV_READ | EV_PERSIST, work_tcp_connection, s);
	event_add(&uc->ev, NULL);
}

int init_streams(void ) {
	struct stream *		s;
	struct transponder *	t = NULL;
	int			sopt;

	/* Check for transponders and find the only one 
	 * By walking the streams open a socket ... */
	s = streams;
	do {

		/* Fill channel pointer from dvbrc */
		if (!s->channel) {
			if (s->cid >= 0) { /* PMT PID */
				s->channel=get_channel_by_id(s->cid);
				if (!s->channel) {
					fprintf(stderr, "ERROR: Unknown channel %u (0x%x)\n", s->cid, s->cid);
					exit(-1);
				}
			} else { /* Special PID (-PID) */
				if (s == streams) {
					fprintf(stderr, "ERROR: Special PIDs (<0, EPG etc.)"
							" must not be the (first or)"
							" last stream entry as they have no"
							" channel assigned!\n");
					exit(-1);
				} else {
					struct pidconsumer * pc;
					s->channel = calloc(sizeof(struct channel), 1);
					s->channel->sid = s->cid;
					s->cid = 0 - s->cid;
					/* Just copy over. */
					s->channel->transponder = streams->channel->transponder;
					s->channel->name = malloc(64);
					sprintf(s->channel->name, "SPECIALPID%02d", s->cid);
					/* Then fill PID table */
					pidtable[s->cid].type = PID_SPECIALPID;
					pc = calloc(sizeof(struct pidconsumer), 1);
					pc->s = s;
					pidtable[s->cid].pc = pc;
					putlog(0, "Setting PID %d type to SPECIALPID for "
					          "PNR %d (%s)", s->cid, s->channel->sid, s->channel->name);
				}
			}
		}

		if (!t) {
			t = s->channel->transponder;
		}

		/* Channels from different transponders ? */
		if (t != s->channel->transponder) {
			fprintf(stderr, "ERROR: The selected channels are from different transponders\n");
			exit(-1);
		}

		memset(&s->ouraddr, 0, sizeof(struct sockaddr_in));

		if (!inet_aton(s->ipaddr, &s->ouraddr.sin_addr)) {
			fprintf(stderr, "ERROR: IP address %s is not valid.\n", s->ipaddr);
			exit(-1);
		}
		s->ouraddr.sin_port=htons(s->port);
		s->ouraddr.sin_family=AF_INET;
		
		if ((ntohl(s->ouraddr.sin_addr.s_addr) >> (4+24)) == (224 >> 4)) {
			/* Multicast address */
			struct ip_mreq		mreq;
			
			s->nuccs = -1;
			s->sockfd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

			sopt = sndbuf_size;
			if (setsockopt(s->sockfd, SOL_SOCKET, SO_SNDBUF, (char *)&sopt, sizeof(sopt))) {
				perror("setsockopt SNDBUF");
			}
			if (connect(s->sockfd, (struct sockaddr *) &s->ouraddr, sizeof(struct sockaddr_in))) {
				perror("connect");
				exit(-1);
			}
			
			/*
			 * Set socket TTL - Be warned - I DoSed a Cisco 5500 RSM by sending
			 * a 60MBit/s stream in 14 Groups all with TTL of 1 and the switch
			 * went to lala land. It seems dropping MCAST traffic is very expensive
			 * in IOS 12.1 land and its even dropped in Layer 3 instead of Layer 2 although
			 * nobody expects ICMP "TTL expired" for MCAST traffic
			 *
			 */
			if (setsockopt(s->sockfd, IPPROTO_IP, IP_MULTICAST_TTL, &s->ttl, sizeof(s->ttl)) != 0) {
				perror("setsockopt MULTICAST_TTL");
			}
			
			/* Join the group (necessary?) */
			memcpy (&mreq.imr_multiaddr, &(s->ouraddr.sin_addr), sizeof(struct in_addr));
			mreq.imr_interface.s_addr = htonl(INADDR_ANY);
			if (setsockopt(s->sockfd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (void*) &mreq, sizeof(mreq)) != 0) {
				perror("sap mcast join");
			}
		} else {
			/* Unicast address. open socket. */
			
			s->nuccs = 0;
		
			s->sockfd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
			
			setfdnonblocking(s->sockfd);
			
			sopt = 1;
			if (setsockopt(s->sockfd, SOL_SOCKET, SO_REUSEADDR, (char *)&sopt, sizeof(sopt))) {
				perror("setsockopt REUSEADDR");
			}
			if (bind(s->sockfd, (struct sockaddr *) &s->ouraddr, sizeof(struct sockaddr_in))) {
				perror("bind");
				exit(-1);
			}
			if (listen(s->sockfd, 5)) {
				perror("listen");
				exit(-1);
			}

			event_set(&s->ev, s->sockfd, EV_READ | EV_PERSIST, new_tcp_connection, s);
			event_add(&s->ev, NULL);
		}
		
		s=s->next;
	} while(s);

	return 1;
}

static void init_sighandlers(void) {
	struct sigaction sia;
	/* If we don't do this, we're likely to die from 'broken pipe'! */
	sia.sa_handler = sigpipehandler;
	sigemptyset(&sia.sa_mask);
	sia.sa_flags = 0;
	sigaction(SIGPIPE, &sia, NULL);
	/* Try a cleaner exit on Ctrl+C */
	sia.sa_handler = siginthandler;
	sigemptyset(&sia.sa_mask);
	sia.sa_flags = 0;
	sigaction(SIGINT, &sia, NULL);
}

static void showhelptxt(char * progname) {
	printf("Syntax: %s -p pnr:ipaddr:port [-d level] [-a adapter] [-h]\n", progname);
	printf("           [-b sndbuf] [-l connectionlimit] [-f file]\n");
	printf("-p nr:ip:port  defines a stream: pid pnr gets streamed to ipaddr:port\n");
	printf("               If you give a unicast address, this creates a HTTP\n");
	printf("               streaming server, if you give a multicast address, this\n");
	printf("               creates a multicast udp stream to the specified multicast\n");
	printf("               address.\n");
	printf("               There can be multiple -p options.\n");
	printf("               Negative pids specify special pids - e.g. -18 for EPG\n");
	printf("-d level       Sets the log level. Higher level gives more output.\n");
	printf("               defaults to 1.\n");
	printf("-a adapter     Selects the adapter for streaming. default 0.\n");
	printf("-b sndbuf      How much bufferspace to request from the OS (SO_SNDBUF)\n");
	printf("-l limit       Limit for number of connections PER STREAM (!).\n");
	printf("               Used for unicast only (there are no connections on mcast)\n");
	printf("-f filename    Read channel list from filename instead of [~/].dvbrc\n");
}

static uint8_t * printtimediff(time_t d) {
	static uint8_t opbuf[100];
	uint8_t secs, mins, hours, days;
	
	secs = d % 60;
	d /= 60;
	mins = d % 60;
	d /= 60;
	hours = d % 24;
	d /= 24;
	days = d;
	sprintf(opbuf, "%3d+%02d:%02d:%02d", days, hours, mins, secs);
	return &opbuf[0];
}

static uint8_t * printtrafcounter(unsigned long long bc) {
	static uint8_t opbuf[100];
	uint8_t tmpunit = ' ';
	
	if (bc >= 100000ULL) {
		tmpunit = 'K';
		bc /= 1000ULL;
	}
	if (bc >= 100000ULL) {
		tmpunit = 'M';
		bc /= 1000ULL;
	}
	if (bc >= 100000ULL) {
		tmpunit = 'G';
		bc /= 1000ULL;
	}
	sprintf(opbuf, "%9llu%c", bc, tmpunit);
	return &opbuf[0];
}

static void doshowstatus(void) {
	struct stream * s = streams;
	unsigned long totclients = 0;
	unsigned long nstreams = 0;
	unsigned long nucstreams = 0;
	
	printf("General Server Status:\n");
	while (s) {
		if (s->nuccs >= 0) {
			totclients += s->nuccs;
			nucstreams++;
		}
		nstreams++;
		s = s->next;
	}
	printf("%lu streams, of which %lu are unicast, and %lu multicast\n",
	       nstreams, nucstreams, (nstreams - nucstreams));
	printf("%lu unicast client%s connected\n", totclients, ((totclients == 1) ? "" : "s"));
#ifdef COUNTSTATS
	printf("%20llu Bytes (%s) received from the network\n",
	       (unsigned long long)totrxbytes,
	       printtrafcounter(totrxbytes));
	printf("%20llu Bytes (%s) sent to the network\n",
	       (unsigned long long)tottxbytes,
	       printtrafcounter(tottxbytes));
	printf("%20llu Bytes (%s) read from the dvb adapter (avg: %5.1lf MBit)\n",
	       (unsigned long long)totbytesfromadapter,
	       printtrafcounter(totbytesfromadapter),
	       (double)(totbytesfromadapter/(((double)(time(NULL) - startupts)) * (1000.0 / 8.0) * 1000.0)));
	printf("%lu successful connections, %lu connection attempts (%5.1lf%%)\n",
	       totconsuccesss, totconattempts,
	       (double)((double)(totconsuccesss * 100) / (double)totconattempts));
#else
	printf("Not collecting any traffic statistics (recompile with -DCOUNTSTATS to see them)\n");
#endif /* COUNTSTATS */
	printf("Uptime: %s\n", printtimediff(time(NULL) - startupts));
}

static void doshowstreams(void) {
	struct stream * s = streams;
	struct unicast_client * uc;
	unsigned long totclients = 0;
	unsigned long nstreams = 0;
	unsigned long nucstreams = 0;

	printf("%-2s %-5s %-2s %-15s %-5s %10s %10s %-8s %-8s %-7s %s\n",
	       "#", "FD", "MU", "IP", "Port", "TX", "RX", "Con.Att.", "Con.Suc.", "Clients", "Stream");
	while (s) {
		printf("%02lu %05d %2s %-15s %5u ",
		       nstreams, s->sockfd, ((s->nuccs >= 0) ? "UC" : "MC"),
		       inet_ntoa(s->ouraddr.sin_addr),
		       ntohs(s->ouraddr.sin_port));
#ifdef COUNTSTATS
		printf("%s ", printtrafcounter(s->txbytes));
		printf("%s ", printtrafcounter(s->rxbytes));
#else
		printf("%10s %10s ", "N/A", "N/A");
#endif /* COUNTSTATS */
		if (s->nuccs >= 0) { /* Unicast */
			unsigned int clisonstream = 0;
			nucstreams++;
			if (s->nuccs > 0) {
				uc = s->ucclients;
			} else {
				uc = NULL;
			}
			while (uc) {
				uc = uc->next;
				totclients++;
				clisonstream++;
			}
#ifdef COUNTSTATS
			printf("%8lu %8lu %3u/%3u ",
			       s->conattempts, s->consuccesss,
			       clisonstream, maxuccs);
#else
			printf("%8s %8s %7s ", "N/A", "N/A", "N/A");
#endif /* COUNTSTATS */
		} else { /* Multicast */
			printf("%8s %8s %7s ", "N/A", "N/A", "N/A");
		}
		printf("%s\n", s->channel->name);
		nstreams++;
		s = s->next;
	}
	printf("--- End of list - %lu stream%s total, %lu unicast\n",
	       nstreams, ((nstreams == 1) ? "" : "s"), nucstreams);
}

static void doshowstreampids(void) {
	struct stream * s = streams;
	unsigned long nstreams = 0;
	int i;
	printf("%-2s", "#");
	for (i = 1; i < N_PIDTYPES; i++) {
		printf(" %7s", pidtdescr[i]);
	}
	printf(" %s\n", "Name");
	while (s) {
		printf("%02lu", nstreams);
		for (i = 1; i < N_PIDTYPES; i++) {
			printf(" %7u", s->pidcounters[i]);
		}
		printf(" %s\n", s->channel->name);
		nstreams++;
		s = s->next;
	}
}

static void doshowclients(void) {
	struct stream * s = streams;
	struct unicast_client * uc;
	unsigned long totclients = 0;
	
	printf("%-5s %-15s %-5s %10s %10s %s\n", "FD", "IP", "OvFl.", "TX", "RX", "Stream");
	while (s) {
		if (s->nuccs > 0) { /* Unicast, at least one client */
			uc = s->ucclients;
			while (uc) {
				printf("%05d %-15s %5lu ",
				       uc->sockfd,
				       inet_ntoa(uc->remaddr.sin_addr),
				       uc->sendoverflows);
#ifdef COUNTSTATS
				printf("%s ", printtrafcounter(uc->txbytes));
				printf("%s ", printtrafcounter(uc->rxbytes));
#else
				printf("%10s %10s ", "N/A", "N/A");
#endif /* COUNTSTATS */
				printf("%s%s\n",
				       (uc->hadrequest ? "" : "!"),
				       s->channel->name);
				totclients++;
				uc = uc->next;
			}
		}
		s = s->next;
	}
	printf("--- End of list - %lu unicast client%s total\n",
	       totclients, ((totclients == 1) ? "" : "s"));
}

unsigned char * mytabcompleter(unsigned char * curinp) {
	return NULL;
}

void handleterminput(int fd, short event, void * up) {
	char buf[10];
	
	if (event & EV_READ) {
		int rdres;
		rdres = read(fd, buf, 1);
		if (rdres != 1) {
			return;
		}
		if (addtoinputbuf(buf[0])) {
			unsigned char * kbdcmd;
			
			kbdcmd = getinputline();
			putlog(0, "[Console command]: %s", kbdcmd);
			jmptocol1();
			cleartoeol();
			if        (strcmp(kbdcmd, "status") == 0) {
				doshowstatus();
				clearinputbuf();
			} else if (strcmp(kbdcmd, "showclients") == 0) {
				doshowclients();
				clearinputbuf();
			} else if (strcmp(kbdcmd, "showstreams") == 0) {
				doshowstreams();
				clearinputbuf();
			} else if (strcmp(kbdcmd, "showstreampids") == 0) {
				doshowstreampids();
				clearinputbuf();
			} else if (strcmp(kbdcmd, "quit") == 0) {
				clearinputbuf();
				printf("Goodbye cruel world.\n");
				putlog(0, "--- QUITTING through quit command from console ---");
				jmptocol1();
				cleartoeol();
				exit(0);
			} else if (strcmp(kbdcmd, "help") == 0) {
				printf("Available commands:\n");
				printf("help           shows this text\n");
				printf("quit           quits\n");
				printf("showclients    lists current (unicast) clients\n");
				printf("showstreams    lists streams\n");
				printf("showstreampids shows PIDs belonging to the streams\n");
				printf("status         shows general status\n");
				clearinputbuf();
			} else {
				printf("Unknown command.\n");
			}
			redrawinputline();
		}
	}
}

int main(int argc, char **argv) {
	int		adapter=0;
	char		ch;
	char *		chanconffile = NULL;
	static struct event	termevent;

	printf("Getstream Version %s%s Compiled on %s %s\n",
	       "20051129", VERSIONSUF, __DATE__, __TIME__);
	
	while ((ch = getopt(argc, argv, "a:b:d:f:l:p:h")) != -1) {
		switch (ch) {
			case 'a':
				adapter = strtol(optarg, NULL, 10);
				break;
			case 'b':
				sndbuf_size = strtoul(optarg, NULL, 10);
				break;
			case 'd':
				console_loglevel = strtol(optarg, NULL, 10);
				break;
			case 'f':
				chanconffile = optarg;
				break;
			case 'h':
				showhelptxt(argv[0]);
				exit(-1);
				break;
			case 'l':
				maxuccs = strtoul(optarg, NULL, 10);
				break;
			case 'p':
				parse_cmdline_program(optarg);
				break;
			default:
				fprintf(stderr, "Unknown command line arg\n");
				fprintf(stderr, "Try %s -h for help\n", argv[0]);
				exit(-1);
				break;
		}
	}

	if (!streams) {
		fprintf(stderr, "Need at least one program to stream\n");
		exit(-1);
	}

	putlog(1, "Reading dvbrc...");
	read_dvbrc(chanconffile);

	putlog(5, "installing SIGPIPE/SIG handler...");
	init_sighandlers();
	
	/* Initialize libevent */
	putlog(5, "libevent init...");
	event_init();
	
	/* Initialize terminal */
	putlog(5, "Terminal init...");
	if (setupstdin()) {
		event_set(&termevent, STDIN_FILENO, EV_READ | EV_PERSIST, handleterminput, NULL);
		event_add(&termevent, NULL);
		registertabcompletioncallback(mytabcompleter);
	}

	putlog(1, "Initialize Streams...");
	init_streams();

	/* Tune to the one and only transponder */
	putlog(0, "Tuning...");
	fe_tune(adapter, streams->channel->transponder);

	/* Initialize demux0 dvr0 and co */
	putlog(1, "DVB and sap init...");
	dvb_init(adapter);
	init_sap();

	/* Pigs can fly */
	putlog(0, "running.");
	startupts = time(NULL);
	event_dispatch();

	return 0;
}
