/*
 * $Id: getstream.h,v 1.7 2007/02/10 21:42:48 simimeie Exp $
 */

#ifndef __getstream_h__
#define __getstream_h__

#include <event.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <inttypes.h>

#include "dvbrc.h" /* Needed for type definitions. unclean. */

#define DEBUG

#define MAX_MCAST_PAYLOAD	(1500-40)
#define TS_PACKET_SIZE		188

#define MAXSENDOVERFLOWS	500	/* After this many overflows the connection gets dropped */

#define MAX_PID			0x1fff

enum {
	PID_NONE = 0,
	PID_PMT,
	PID_PCR,
	PID_VIDEO,
	PID_AUDIO,
	PID_PRIVATE,
	PID_USER,
	PID_OTHER,
	PID_SPECIALPID,
	N_PIDTYPES
};

static const char * const pidtdescr[] = {
	"none", "PMT", "PCR", "VIDEO", "AUDIO",
       	"PRIVATE", "USER", "OTHER", "SPECIAL", "BUG_MUST_NOT_PRINT_THIS_BUG"
};

struct unicast_client {
	struct unicast_client *	next;
	int			sockfd;
	struct event		ev;
	unsigned long		sendoverflows;
	char			hadrequest;
	struct sockaddr_in	remaddr;
#ifdef COUNTSTATS
	uint64_t		rxbytes;
	uint64_t		txbytes;
#endif /* COUNTSTATS */
};

struct stream {
	struct stream *		next;
	struct channel *	channel;
	struct unicast_client *	ucclients;
	int			nuccs; /* Number of UniCast Clients; -1 = multicast */
	struct event		ev; /* Only needed for unicast */
	int			cid;
	int			sockfd;
	int			pbufvalid;
	unsigned char 		pbuf[MAX_MCAST_PAYLOAD];
	char			ipaddr[128];
	struct sockaddr_in	ouraddr;
	int			port;
	int			ttl;
	unsigned int		pidcounters[N_PIDTYPES];
	char			pid0cc;
#ifdef COUNTSTATS
	uint64_t		rxbytes; /* Bytes read from the network */
	uint64_t		txbytes; /* Bytes sent to the network */
	unsigned long		conattempts; /* Connection attempts (only unicast) */
	unsigned long		consuccesss; /* Successful connections */
#endif /* COUNTSTATS */
};

struct pidconsumer {
	struct stream *		s;
	struct pidconsumer *	next;
};

struct pidentry {
	unsigned char		type;
	unsigned char		cc;
	time_t			lastvalid;
	struct pidconsumer *	pc;
};

extern struct stream *		streams;
extern struct pidentry		pidtable[MAX_PID];

extern unsigned int		maxuccs; /* Maximum number of unicast clients */
#ifdef COUNTSTATS
extern uint64_t			totrxbytes; /* Total network counters */
extern uint64_t			tottxbytes;
extern unsigned long		totconattempts; /* Connection attempts (only unicast) */
extern unsigned long		totconsuccesss; /* Successful connections */
extern uint64_t			totbytesfromadapter;
#endif /* COUNTSTATS */

/* Some functions */
void drop_tcp_connection(struct stream * s, int fd);
void setfdnonblocking(int fd);
void send_ts_packet(struct stream * s, uint8_t * tsp);

#endif /* __getstream_h__ */
